#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "raichu_alu.c"
#include "raichu_fpu.c"

#define REGNUM ((unsigned int)(128))
#define MEMSIZE ((unsigned int)(1024*1024)) /* MEMSIZE must be 2^n */

struct resources{
  unsigned int pc, ipc;
  unsigned int ir;
  unsigned int reg[REGNUM];
  unsigned int mem[MEMSIZE];
};

struct configs{
  FILE *fpi, *fpo, *fpl;
  char binname[1024];
  unsigned char bp[MEMSIZE];
  unsigned char logrflg[REGNUM];
  unsigned char logmflg[MEMSIZE];
};

char buf[1024];

struct configs *new_configs(char *filename){
  FILE *fp;
  struct configs *ncfg;
  
  if((fp = fopen(filename, "r"))){
    int i;
    
    ncfg = (struct configs *)malloc(sizeof(struct configs));
    if(!ncfg){
      fprintf(stderr, "not enough memory\n");
      fclose(fp);
      return NULL;
    }
    
    if(!fgets(buf, 1024, fp)){
      fprintf(stderr, "file input error\n");
      free(ncfg);
      fclose(fp);
      return NULL;
    }
    for(i = 0; buf[i] != '\n' && buf[i] != '\r' && buf[i] != '\0'; i++);
    buf[i] = '\0';
    memcpy(ncfg->binname, buf, sizeof(char) * (i + 1));
    
    if(!fgets(buf, 1024, fp)){
      fprintf(stderr, "file input error\n");
      free(ncfg);
      fclose(fp);
      return NULL;
    }
    for(i = 0; buf[i] != '\n' && buf[i] != '\r' && buf[i] != '\0'; i++);
    buf[i] = '\0';
    if(!((ncfg->fpi) = fopen(buf, "r"))){
      fprintf(stderr, "input file %s not found\n", buf);
      free(ncfg);
      fclose(fp);
      return NULL;
    }
    
    if(!fgets(buf, 1024, fp)){
      fprintf(stderr, "file input error\n");
      fclose(ncfg->fpi);
      free(ncfg);
      fclose(fp);
      return NULL;
    }
    for(i = 0; buf[i] != '\n' && buf[i] != '\r' && buf[i] != '\0'; i++);
    buf[i] = '\0';
    if(!((ncfg->fpo) = fopen(buf, "wb"))){
      fprintf(stderr, "file %s create error\n", buf);
      fclose(ncfg->fpi);
      free(ncfg);
      fclose(fp);
      return NULL;
    }
    
    if(!fgets(buf, 1024, fp)){
      fprintf(stderr, "file input error\n");
      fclose(ncfg->fpo);
      fclose(ncfg->fpi);
      free(ncfg);
      fclose(fp);
      return NULL;
    }
    for(i = 0; buf[i] != '\n' && buf[i] != '\r' && buf[i] != '\0'; i++);
    buf[i] = '\0';
    if(!((ncfg->fpl) = fopen(buf, "w"))){
      fprintf(stderr, "file %s create error\n", buf);
      fclose(ncfg->fpo);
      fclose(ncfg->fpi);
      free(ncfg);
      fclose(fp);
      return NULL;
    }
    
    for(i = 0; i < REGNUM; i++)
      (ncfg->logrflg[i]) = 0;
    for(i = 0; i < MEMSIZE; i++){
      (ncfg->bp[i]) = 0;
      (ncfg->logmflg[i]) = 0;
    }
    
    while(fgets(buf, 1024, fp)){
      if(buf[0] == 'l' && buf[1] == 'n'){
	unsigned int adr = 0;
	
	for(i = 0; i < 8; i++){
	  if(buf[3 + i] >= '0' && buf[3 + i] <= '9')
	    adr = (adr << 4) + (unsigned int)(buf[3 + i] - '0');
	  else if(buf[3 + i] >= 'a' && buf[3 + i] <= 'f')
	    adr = (adr << 4) + (unsigned int)(buf[3 + i] - 'a' + 10);
	  else{
	    fprintf(stderr, "file input error\n");
	    fclose(ncfg->fpl);
	    fclose(ncfg->fpo);
	    fclose(ncfg->fpi);
	    free(ncfg);
	    fclose(fp);
	    return NULL;
	  }
	}
	ncfg->bp[adr & (MEMSIZE - 1)] = 1;
      }
      else if(buf[0] == 'm'){
	unsigned int adr = 0;
	
	for(i = 0; i < 8; i++){
	  if(buf[2 + i] >= '0' && buf[2 + i] <= '9')
	    adr = (adr << 4) + (unsigned int)(buf[2 + i] - '0');
	  else if(buf[2 + i] >= 'a' && buf[2 + i] <= 'f')
	    adr = (adr << 4) + (unsigned int)(buf[2 + i] - 'a' + 10);
	  else{
	    fprintf(stderr, "file input error\n");
	    fclose(ncfg->fpl);
	    fclose(ncfg->fpo);
	    fclose(ncfg->fpi);
	    free(ncfg);
	    fclose(fp);
	    return NULL;
	  }
	}
	ncfg->logmflg[adr & (MEMSIZE - 1)] = 1;
      }
      if(buf[0] == 'r'){
	unsigned int adr = 0;
	
	for(i = 0; i < 3; i++){
	  if(buf[2 + i] >= '0' && buf[2 + i] <= '9')
	    adr = (adr * 10) + (unsigned int)(buf[2 + i] - '0');
	  else if(i != 0)
	    break;
	  else{
	    fprintf(stderr, "file input error\n");
	    fclose(ncfg->fpl);
	    fclose(ncfg->fpo);
	    fclose(ncfg->fpi);
	    free(ncfg);
	    fclose(fp);
	    return NULL;
	  }
	}
	if(adr >= REGNUM){
	  fprintf(stderr, "file input error\n");
	  fclose(ncfg->fpl);
	  fclose(ncfg->fpo);
	  fclose(ncfg->fpi);
	  free(ncfg);
	  fclose(fp);
	  return NULL;
	}
	ncfg->logrflg[adr] = 1;
      }
    }
    
    if(feof(fp)){
      fclose(fp);
      return ncfg;
    }
    else{
      fprintf(stderr, "file input error\n");
      fclose(ncfg->fpl);
      fclose(ncfg->fpo);
      fclose(ncfg->fpi);
      free(ncfg);
      fclose(fp);
      return NULL;
    }
  }
  else{
    fprintf(stderr, "config file %s not found\n", filename);
    return NULL;
  }
}

void delete_configs(struct configs *cfg){
  fclose(cfg->fpl);
  fclose(cfg->fpo);
  fclose(cfg->fpi);
  free(cfg);
}

struct resources *new_resources(char *filename){
  FILE *fp;
  
  if((fp = fopen(filename, "rb"))){
    struct resources *nrs;
    unsigned int adr;
    
    nrs = (struct resources *)malloc(sizeof(struct resources));
    
    if(!nrs){
      fprintf(stderr, "not enough memory\n");
      fclose(fp);
      return NULL;
    }
    
    for(adr = 0; ;){
      unsigned int readed;
      unsigned int i, j;
      
      readed = fread(buf, sizeof(char), 1024, fp);
      
      if(readed < 1024 && ferror(fp)){
	fprintf(stderr, "file input error\n");
	free(nrs);
	fclose(fp);
	return NULL;
      }
      else if(readed & 3){
	fprintf(stderr, "bin file form error\n");
	free(nrs);
	fclose(fp);
	return NULL;
      }
      
      for(i = 0; i < readed; adr++){
  	unsigned int op = 0;
	
	if(adr >= MEMSIZE){
	  fprintf(stderr, "file too large\n");
	  free(nrs);
	  fclose(fp);
	  return NULL;
	}
	
	for(j = 0; j < 4; j++, i++)
	  op = (op << 8) + ((unsigned char)buf[i]);
	(nrs->mem[adr]) = op;
      }
      
      if(feof(fp))
	break;
    }
    
    for(; adr < MEMSIZE; adr++)
      nrs->mem[adr] = 0x00000000U;
    
    for(adr = 0; adr < REGNUM; adr++)
      nrs->reg[adr] = 0x00000000U;
    nrs->pc = 0x00000001U;
    nrs->ipc = 0x00000000U;
    nrs->ir = nrs->mem[0x00000000U];
    
    fclose(fp);
    
    return nrs;
  }
  else{
    fprintf(stderr, "binary file %s not found\n", filename);
    return NULL;
  }
}

void delete_resources(struct resources *rs){
  free(rs);
}

void print_resources(struct resources *rs, struct configs *cfg){
  unsigned int i;
  
  fprintf(cfg->fpl, "\nnextPC : 0x%08x\n", rs->ipc);
  fprintf(cfg->fpl, "nextnextPC : 0x%08x\n", rs->pc);
  fprintf(cfg->fpl, "nextIR : 0x%08x\n", rs->ir);
  
  for(i = 0; i < REGNUM; i++){
    if(cfg->logrflg[i])
      fprintf(cfg->fpl, "REG[%d] : 0x%08x\n", i, rs->reg[i]);
  }
  
  for(i = 0; i < MEMSIZE; i++){
    if(cfg->logmflg[i])
      fprintf(cfg->fpl, "MEM[0x%08x] : 0x%08x\n", i, rs->mem[i]);
  }
}

int stepexec(struct resources *rs, struct configs *cfg){
  unsigned int inst, pcnt;
  int endflg = 0;
  
  pcnt = rs->ipc;
  inst = rs->ir;
  
  (rs->ipc) = (rs->pc);
  (rs->ir) = (rs->mem[rs->pc]);
  
  (rs->pc) = ((rs->pc) + 1) & (MEMSIZE - 1);
  
  if(inst & 0x80000000U){
    if(inst & 0x40000000U){/* jump / branch */
      if(inst & 0x00600000U){/* jump */
	if(inst & 0x00400000U){
	  unsigned int in1 = ((inst & 0x3f800000U) >> 4) | (inst & 0x0007ffffU);
	  in1 = in1 | (0x00000000U - (in1 & 0x02000000U));
	  
	  switch(inst & 0x00380000U){
	  case 0x00200000U:/* jmp imm */
	    (rs->pc) = ((pcnt + in1) & (MEMSIZE - 1));
	    break;
	  case 0x00280000U:/* jmpl imm */
	    (rs->reg[127]) = (rs->pc);
	    (rs->pc) = ((pcnt + in1) & (MEMSIZE - 1));
	    break;
	  default:
	    fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	  }
	}
	else{
	  switch(inst & 0x001c0000U){
	  case 0x00000000U:/* jmpr r1 */
	    (rs->pc) = ((rs->reg[(inst >> 11) & 127]) & (MEMSIZE - 1));
	    break;
	  case 0x00080000U:/* jmprl r1 */
	    (rs->reg[127]) = (rs->pc);
	    (rs->pc) = ((rs->reg[(inst >> 11) & 127]) & (MEMSIZE - 1));
	    break;
	  case 0x00100000U:/* ret */
	    (rs->pc) = ((rs->reg[127]) & (MEMSIZE - 1));
	    break;
	  case 0x00180000U:/* end */
	  case 0x001c0000U:/* end */
	    (rs->pc) = pcnt;
	    endflg = 1;
	    break;
	  default:
	    fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	  }
	}
      }
      else{
	unsigned int *r1 = &(rs->reg[(inst >> 23) & 127]);
	unsigned int in1 = (inst & 0x0007ffffU);
	
	in1 = in1 | (0x00000000U - (in1 & 0x00040000U));
	
	switch(inst & 0x00180000U){
	case 0x00000000U:/* beq r1 imm */
	  if((*r1) == 0)
	    (rs->pc) = ((pcnt + in1) & (MEMSIZE - 1));
	  break;
	case 0x00080000U:/* bne r1 imm */
	  if((*r1) != 0)
	    (rs->pc) = ((pcnt + in1) & (MEMSIZE - 1));
	  break;
	case 0x00100000U:/* bgt r1 imm */
	  if((*r1) > 0)
	    (rs->pc) = ((pcnt + in1) & (MEMSIZE - 1));
	  break;
	case 0x00180000U:/* bge r1 imm */
	  if((*r1) >= 0)
	    (rs->pc) = ((pcnt + in1) & (MEMSIZE - 1));
	  break;
	}
      }
    }
    else{/* memory access / IO / immediate */
      unsigned int *r1 = &(rs->reg[(inst >> 23) & 127]);
      
      if(inst & 0x00600000U){
	unsigned int in1 = (rs->reg[(inst >> 11) & 127]);
	
	if(inst & 0x00040000U){/* memory/IO with imm */
	  unsigned int in2 = (inst & 2047) | (0x00000000U - (inst & 1024));
	  
	  switch(inst & 0x00780000U){
	  case 0x00480000U:/* loadi r1 r2 imm */
	    *r1 = (rs->mem[(in1 + in2) & (MEMSIZE - 1)]);
	    break;
	  case 0x00680000U:/* storei r1 r2 imm */
	    (rs->mem[(in1 + in2) & (MEMSIZE - 1)]) = (*r1);
	    break;
	  case 0x00500000U:/* in r1 imm */
	    {
	      (*r1) = 0x00000000U;
	      
	      if(!fgets(buf, 1024, cfg->fpi)){
		if(feof(cfg->fpi))
		  fprintf(stderr, "no input\n");
		else
		  fprintf(stderr, "input error\n");
	      }
	      else{
		int i;
		
		for(i = 0; i < 8; i++){
		  if(buf[i] >= '0' && buf[i] <= '9')
		    (*r1) = ((*r1) << 4) + (unsigned int)(buf[i] - '0');
		  else if(buf[i] >= 'a' && buf[i] <= 'f')
		    (*r1) = ((*r1) << 4) + (unsigned int)(buf[i] - 'a' + 10);
		  else{
		    fprintf(stderr, "input form error");
		    break;
		  }
		}
	      }
	    }
	    break;
	  case 0x00700000U:{/* out r1 imm */
	    char c = (char)((unsigned char)(rs->reg[(inst >> 11) & 127]));
		fwrite(&c, sizeof(char), 1, cfg->fpo);
	     /*fprintf(cfg->fpo, "%08x\n", rs->reg[(inst >> 11) & 127]));*/
		break;
	  }
	  default:
	    fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	  }
	}
	else{/* memory/IO with register */
	  unsigned int in2 = (rs->reg[(inst >> 4) & 127]);
	  
	  if(inst & 0x0000000fU)
	    fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	  
	  switch(inst & 0x00780000U){
	  case 0x00400000U:/* load r1 r2 */
	    fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	    /**r1 = (rs->mem[in1 & (MEMSIZE - 1)]);*/
	    break;
	  case 0x00600000U:/* store r1 r2 */
	    fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	    /*(rs->mem[in1 & (MEMSIZE - 1)]) = (*r1);*/
	    break;
	  case 0x00480000U:/* loadd r1 r2 r3 */
	    *r1 = (rs->mem[(in1 + in2) & (MEMSIZE - 1)]);
	    break;
	  case 0x00680000U:/* stored r1 r2 r3 */
	    (rs->mem[(in1 + in2) & (MEMSIZE - 1)]) = (*r1);
	    break;
	  default:
	    fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	  }
	}
      }
      else{/* hi r1 imm */
	*r1 = (inst & 0x001fffffU) << 11;
      }
    }
  }
  else{
    if(inst & 0x40000000U){/* FPU */
      unsigned int *r1 = &(rs->reg[(inst >> 23) & 127]);
      unsigned int in1 = (rs->reg[(inst >> 11) & 127]);
      
      if(inst & 0x00400000U){/* ftoi/itof */
	if(inst & 0x0000000fU)
	  fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	
	switch(inst & 0x003c0000U){
	case 0x00000000U:/* itof r1 r2 */
	  *r1 = fpu_itof(in1);
	  break;
	case 0x00200000U:/* ftoi r1 r2 */
	  *r1 = fpu_ftoi(in1);
	  break;
	default:
	  fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	}
      }
      else{
	unsigned int in2 = (rs->reg[(inst >> 4) & 127]);
	
	if(inst & 0x0000000fU)
	  fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	
	switch(inst & 0x003c0000U){
	case 0x00000000U:/* fadd r1 r2 r3 */
	  *r1 = fpu_fadd(in1, in2);
	  break;
	case 0x00080000U:/* fsub r1 r2 r3 */
	  *r1 = fpu_fsub(in1, in2);
	  break;
	case 0x00100000U:/* fmul r1 r2 r3 */
	  *r1 = fpu_fmul(in1, in2);
	  break;
	case 0x00180000U:/* fdiv r1 r2 r3 */
	  fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	  /**r1 = fpu_fdiv(in1, in2);*/
	  break;
	case 0x00200000U:/* sqrt r1 r2 */
	  fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	  /**r1 = fpu_sqrt(in1);*/
	  break;
	case 0x00280000U:/* finvsqrt r1 r2 */
	  *r1 = fpu_finvsqrt(in1);
	  break;
	case 0x00300000U:/* fless r1 r2 e3 */
	  *r1 = fpu_fless(in1, in2);
	  break;
	default:
	  fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	}
      }
    }
    else{/* ALU */
      unsigned int *r1 = &(rs->reg[(inst >> 23) & 127]);
      unsigned int in1 = (rs->reg[(inst >> 11) & 127]);
      
      if(inst & 0x00400000U){
	if(inst & 0x00200000U){/* shift */
	  if(inst & 0x00040000U){/* shift with imm */
	    unsigned int in2 = (inst & 63);
	    
	    switch(inst & 0x00180000U){
	    case 0x00000000U:/* slli r1 r2 imm */
	      *r1 = alu_sll(in1, in2);
	      break;
	    case 0x00080000U:/* srli r1 r2 imm */
	      *r1 = alu_srl(in1, in2);
	      break;
	    case 0x00100000U:/* srai r1 r2 imm */
	      *r1 = alu_sra(in1, in2);
	      break;
	    default:
	      fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	    }
	  }
	  else{/* shift with register */
	    unsigned int in2 = (rs->reg[(inst >> 4) & 127]) & 63;
	    
	    if(inst & 0x0000000fU)
	      fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	    
	    switch(inst & 0x00180000U){
	    case 0x00000000U:/* sll r1 r2 r3 */
	      *r1 = alu_sll(in1, in2);
	      break;
	    case 0x00080000U:/* srl r1 r2 r3 */
	      *r1 = alu_srl(in1, in2);
	      break;
	    case 0x00100000U:/* sra r1 r2 r3 */
	      *r1 = alu_sra(in1, in2);
	      break;
	    default:
	      fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	    }
	  }
	}
	else{/* logic */
	  if(inst & 0x00040000U){/* logic with imm */
	    unsigned int in2 = (inst & 2047);
	    
	    switch(inst & 0x00180000U){
	    case 0x00000000U:/* andi r1 r2 imm */
	      *r1 = alu_and(in1, in2);
	      break;
	    case 0x00080000U:/* ori r1 r2 imm */
	      *r1 = alu_or(in1, in2);
	      break;
	    case 0x00100000U:/* xori r1 r2 imm */
	      *r1 = alu_xor(in1, in2);
	      break;
	    default:
	      fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	    }
	  }
	  else{/* logic with register */
	    unsigned int in2 = (rs->reg[(inst >> 4) & 127]);
	    
	    if(inst & 0x0000000fU)
	      fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	    
	    switch(inst & 0x00180000U){
	    case 0x00000000U:/* and r1 r2 r3 */
	      *r1 = alu_and(in1, in2);
	      break;
	    case 0x00080000U:/* or r1 r2 r3 */
	      *r1 = alu_or(in1, in2);
	      break;
	    case 0x00100000U:/* xor r1 r2 r3 */
	      *r1 = alu_xor(in1, in2);
	      break;
	    case 0x00180000U:/* not r1 r2 r3 */
	      *r1 = alu_not(in1);
	      break;
	    default:
	      fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	    }
	  }
	}
      }
      else{/* arith/cmp */
	if(inst & 0x00040000U){/* arith/cmp with imm */
	  unsigned int in2 = (inst & 2047) | (0x00000000U - (inst & 1024));
	  
	  switch(inst & 0x00380000U){
	  case 0x00000000U:/* addi r1 r2 imm */
	    *r1 = alu_add(in1, in2);
	    break;
	  case 0x00100000U:/* muli r1 r2 imm */
	    fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	    /**r1 = alu_mul(in1, in2);*/
	    break;
	  case 0x00280000U:/* cmpeqi r1 r2 imm */
	    *r1 = alu_cmpeq(in1, in2);
	    break;
	  case 0x00300000U:/* cmplti r1 r2 imm */
	    *r1 = alu_cmplt(in1, in2);
	    break;
	  case 0x00380000U:/* cmpgti r1 r2 imm */
	    *r1 = alu_cmplt(in2, in1);
	    break;
	  default:
	    fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	  }
	}
	else{/* arith/cmp with register */
	  unsigned int in2 = (rs->reg[(inst >> 4) & 127]);
	  
	  if(inst & 0x0000000fU)
	    fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	  
	  switch(inst & 0x00380000U){
	  case 0x00000000U:/* add r1 r2 r3 */
	    *r1 = alu_add(in1, in2);
	    break;
	  case 0x00080000U:/* sub r1 r2 r3 */
	    *r1 = alu_sub(in1, in2);
	    break;
	  case 0x00100000U:/* mul r1 r2 r3 */
	    *r1 = alu_mul(in1, in2);
	    break;
	  case 0x00180000U:/* mulh r1 r2 r3 */
	    fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	    /**r1 = alu_mulh(in1, in2);*/
	    break;
	  case 0x00200000U:/* div r1 r2 r3 */
	    *r1 = alu_div(in1, in2);
	    break;
	  case 0x00280000U:/* cmpeq r1 r2 r3 */
	    *r1 = alu_cmpeq(in1, in2);
	    break;
	  case 0x00300000U:/* cmplt r1 r2 r3 */
	    *r1 = alu_cmplt(in1, in2);
	    break;
	  default:
	    fprintf(stderr, "operation 0x%08x not allowed\n", inst);
	  }
	}
      }
    }
  }
  
  (rs->reg[0]) = 0x00000000U;
  
  return endflg;
}

int main(int argc, char *argv[]){
  struct resources *rs;
  struct configs *cfg;
  unsigned int spmin = 0x00100000, spresetflg = 0;
  unsigned int totalinumh = 0, totalinuml = 0;
  unsigned int inumh = 0, inuml = 0;
  
  if(argc < 2){
    fprintf(stderr, "usage: raichu [program file name]\n");
    return 1;
  }
  
  if(sizeof(unsigned int) != 4){
    fprintf(stderr, "sizeof(unsigned int) must be 4\n");
    return 1;
  }
  
  if(((0x80000001U >> 1) != 0x40000000U) || 
     ((0x40000000U << 1) != 0x80000000U) || 
     ((0x80000000U << 1) != 0x00000000U)){
    fprintf(stderr, "sll and srl needed\n");
    return 1;
  }
  
  if(((-1) >> 1) != -1){
    fprintf(stderr, "sra needed\n");
    return 1;
  }
  
  cfg = new_configs(argv[1]);
  if(!cfg)
    return 1;
  
  rs = new_resources(cfg->binname);
  if(!rs){
    delete_configs(cfg);
    return 1;
  }
  
  while(1){
    int endflg;
    
    if((rs->ir) == 0x00000000U){
      inuml++;
      if(!inuml)
	inumh++;
    }
    
    endflg = stepexec(rs, cfg);
    
    if(cfg->bp[rs->ipc]){
      print_resources(rs, cfg);
    }
    
    if((rs->reg[125]) > 0)
      spresetflg = 1;
    if(((0x00100000 - spmin) < (0x00100000 - (rs->reg[125]))) && spresetflg)
      spmin = (rs->reg[125]);
    totalinuml++;
    if(!totalinuml)
      totalinumh++;
    
    if(endflg)
      break;
  }
  
  printf("%u * 2^32 + %u instructions\n", totalinumh, totalinuml);
  printf("%u * 2^32 + %u nops\n", inumh, inuml);
  printf("sp >= %08x\n", spmin);
  printf("hp <= %08x\n", rs->reg[126]);
  
  delete_resources(rs);
  delete_configs(cfg);
  
  return 0;
}
